(*  Title     : A Shallow Encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Old vs new definition of Chopstar\<close> 

theory AltChopstarSem
imports Semantics 
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="
  cvc4 z3 e spass vampire
"]
*)

text \<open>
 We show that the old and new definition of chopstar are the same.
 
\<close>

subsection \<open>Definition\<close>

definition chopstar_d_old :: "('a::world) formula \<Rightarrow> 'a formula"
where "chopstar_d_old F \<equiv> \<lambda>s. (\<exists> (l::index). index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen s) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                               ((sub (nth l i) (nth l (i+1)) s) \<Turnstile> F)
                           )
                          )"

syntax
 "_chopstar_d_old"     :: "lift \<Rightarrow> lift"        ("(chopstarold _)" [85] 85)

syntax (ASCII)
 "_chopstar_d_old"     :: "lift \<Rightarrow> lift"        ("(chopstarold _)" [85] 85)

translations
 "_chopstar_d_old"    \<rightleftharpoons> "CONST chopstar_d_old"


subsection \<open>Lemmas\<close>

lemma chopstar_help_1:
  "( \<exists>l. l= \<langle>0\<rangle> \<and> index_sequence 0 l \<and> 
         Interval.nth l (intlen l) = (intlen \<sigma>) \<and> 
         (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                               ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
                           )) \<longleftrightarrow> (intlen \<sigma> =0) " 
by (simp add: index_sequence_def)

lemma chopstar_help_2:
 "(\<forall>i. (0<i \<and> i< 1+(intlen ls)) \<longrightarrow>  
                               ((sub (nth ls (i-1)) (nth ls ((i-1)+1)) \<sigma>) \<Turnstile> f)
                           ) =
  (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow>  
                               ((sub (nth ls (i)) (nth ls ((i)+1)) \<sigma>) \<Turnstile> f)
                    )"
by (metis Suc_eq_plus1 Suc_pred add_diff_cancel_right' add_less_cancel_left 
          add_nonneg_pos le_add2 le_add_same_cancel2 plus_1_eq_Suc zero_less_one)

lemma chop_power_chain:
  "(\<exists> (l::index). (intlen l) = (Suc n) \<and> index_sequence 0 l \<and> (nth l (intlen l)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                               ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
                           )
                ) =
   (\<exists> k. 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls. (intlen ls) = n \<and>index_sequence 0 (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) 
                 \<and> (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub (nth ls (i)) (nth ls ((i)+1)) (suffix k \<sigma>)) \<Turnstile> f)
                    ))
         ) "
proof -
 have " (\<exists> (l::index). (intlen l) = (Suc n) \<and> index_sequence 0 l \<and> 
                       (nth l (intlen l)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                               ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
                           )
                )
      = 
      (\<exists> x ls l.  (intlen l) = (Suc n) \<and> l=x\<odot>ls \<and> index_sequence 0 l \<and> 
                  (nth l (intlen l)) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                               ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
                           )
               )" 
 by (metis interval_intlen_cons_1 zero_less_Suc)
 also have " ... =
            (\<exists> x ls l.  (intlen l) = (Suc n) \<and> l=x\<odot>ls \<and> index_sequence 0 (x\<odot>ls) \<and> 
                          (nth (x\<odot>ls) (intlen (x\<odot>ls))) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen (x\<odot>ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f)
                           )
                     )" 
 by auto
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> index_sequence 0 (x\<odot>ls) \<and> 
                        (nth (x\<odot>ls) (intlen (x\<odot>ls))) = (intlen \<sigma>) \<and>
                           (\<forall>i. (0\<le>i \<and> i< (intlen (x\<odot>ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f)
                           )
                  )" 
 by auto
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence 0 (x \<odot> ls) \<and> 
                        (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and>
                           ((\<forall>i. (0\<le>i \<and> i< (intlen (x\<odot>ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f))
                           )
                 )" 
 by (simp add: index_sequence_def)
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                        (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and>
                       (x < (nth ls 0) \<and> 
                       ((\<forall>i. (0\<le>i \<and> i< (intlen (x\<odot>ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f))
                           )
                       )
                   )" 
 using interval_idx_cons by auto
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                        (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and>
                       (x < (nth ls 0) \<and> 
                        ((sub (nth (x\<odot>ls) 0) (nth (x\<odot>ls) (1)) \<sigma>) \<Turnstile> f)
                       \<and>
                       ((\<forall>i. (0<i \<and> i< 1+ (intlen (ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f))
                           )
                       )
                   )"
 by (metis (no_types, lifting) One_nat_def add.right_neutral add_Suc add_Suc_right 
           add_cancel_right_left interval_intlen_cons not_gr_zero zero_le zero_less_Suc)
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and>
                       (x < (nth ls 0) \<and> (nth (x\<odot>ls) 0) = x \<and> (nth (x\<odot>ls) (1)) = (nth ls 0) \<and>
                        ((sub (nth (x\<odot>ls) 0) (nth (x\<odot>ls) (1)) \<sigma>) \<Turnstile> f)
                       \<and>
                       ((\<forall>i. (0<i \<and> i< 1+ (intlen (ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f))
                           )
                       )
                   )"  
 by auto
 also have " ... = 
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                          (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and>
                       (x < (nth ls 0) \<and> (nth (x\<odot>ls) 0) = x \<and> (nth (x\<odot>ls) (1)) = (nth ls 0) \<and>
                        ((sub x (nth ls 0) \<sigma>) \<Turnstile> f)
                       \<and>
                       ((\<forall>i. (0<i \<and> i< 1+ (intlen (ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f))
                           )
                       )
                   )"
 by auto
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and>
                       (x < (nth ls 0) \<and> 
                        ((sub x (nth ls 0) \<sigma>) \<Turnstile> f)
                       \<and>
                       ((\<forall>i. (0<i \<and> i< 1+ (intlen (ls))) \<longrightarrow> 
                               ((sub (nth (x\<odot>ls) i) (nth (x\<odot>ls) (i+1)) \<sigma>) \<Turnstile> f))
                           )
                       )
                   )" 
 by auto
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and> 
                 ( x < (nth ls 0) \<and> 
                 ((sub x (nth ls 0) \<sigma>) \<Turnstile> f)
                 \<and>
                  (\<forall>i. (0<i \<and> i< 1+(intlen ls)) \<longrightarrow>  
                               ((sub (nth ls (i-1)) (nth ls ((i-1)+1)) \<sigma>) \<Turnstile> f)
                           )))" 
 using interval_nth_cons by metis
 also have " ... =
            (\<exists> x ls . (intlen ls) = n \<and> x =0 \<and> index_sequence (nth ls 0) (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and> 
                 (x < (nth ls 0) \<and> 
                 ((sub x (nth ls 0) \<sigma>) \<Turnstile> f))
                 \<and> (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow>  
                               ((sub (nth ls (i)) (nth ls ((i)+1)) \<sigma>) \<Turnstile> f)
                    )
               )"
 using chopstar_help_2 by (metis (mono_tags))
 also have " ... =
            (\<exists>  ls . (intlen ls) = n \<and> index_sequence (nth ls 0) (ls) \<and>
                     (nth (ls) (intlen (ls))) = (intlen \<sigma>) \<and> 
                 (0 < (nth ls 0) \<and> 
                 ((sub 0 (nth ls 0) \<sigma>) \<Turnstile> f))
                 \<and> (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub (nth ls (i)) (nth ls ((i)+1)) \<sigma>) \<Turnstile> f)
                    )
               )"
 by simp
 also have " ... =
            (\<exists>  lsk .  (intlen lsk) = n \<and> (nth lsk 0) \<le> intlen \<sigma> \<and> (nth lsk 0) > 0 \<and>
                    ((sub 0 (nth lsk 0) \<sigma>) \<Turnstile> f) \<and>
                    index_sequence (nth lsk 0) (lsk) \<and>
                    (nth (lsk) (intlen (lsk))) = (intlen \<sigma>) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub (nth lsk (i)) (nth lsk ((i)+1)) \<sigma>) \<Turnstile> f)
                    )
               )"
 by (metis Suc_eq_plus1 Suc_pred add.left_neutral eq_iff interval_idx_less_last 
           interval_intlen_gr_zero le_neq_implies_less lessI less_imp_le_nat)
 also have " ... =
            (\<exists> k lsk.  (intlen lsk) = n \<and> (nth lsk 0) \<le> intlen \<sigma> \<and> 
                       (nth lsk 0) > 0 \<and> k=(nth lsk 0) \<and>
                    (sub 0 (nth lsk 0) \<sigma> \<Turnstile> f) \<and> 
                    index_sequence (nth lsk 0) (lsk) \<and> 
                    (nth (lsk) (intlen (lsk))) = (intlen (\<sigma>)) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub ((nth lsk (i))) ((nth lsk ((i)+1))) (\<sigma>)) \<Turnstile> f)
                    )
         )"
 by auto
 also have " ... =
            (\<exists> k lsk.  (intlen lsk) = n \<and> 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and> k=(nth lsk 0) \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (index_sequence k (lsk) \<and> 
                          (nth (lsk) (intlen (lsk))) = ((intlen (suffix k \<sigma>))+k) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub ((nth lsk (i))) ((nth lsk ((i)+1))) (\<sigma>)) \<Turnstile> f)
                    ))
         )" 
 by (simp add: interval_prefix_suffix_intlen interval_suffix_length interval_prefix_length, auto)
 also have " ... =
            (\<exists> k lsk. (intlen lsk) = n \<and>  0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (index_sequence k (lsk) \<and> 
                          (nth (lsk) (intlen (lsk))) = ((intlen (suffix k \<sigma>))+k) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub ((nth lsk (i))) ((nth lsk ((i)+1))) (\<sigma>)) \<Turnstile> f)
                    ))
         ) "
 using index_sequence_def by auto
 also have " ... =
            (\<exists> k.   0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls lsk. (intlen lsk) = n \<and> index_sequence k (lsk) \<and>
                           ls = map (shiftm k) lsk \<and> 
                          (nth (lsk) (intlen (lsk))) = ((intlen (suffix k \<sigma>))+k) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub ((nth lsk (i))) ((nth lsk ((i)+1))) (\<sigma>)) \<Turnstile> f)
                    ))
         )"
 by blast
 also have " ... =
            (\<exists> k.   0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls lsk. (intlen lsk) = n \<and> index_sequence k (lsk) \<and>
                           ls = map (shiftm k) lsk \<and> 
                           index_sequence 0 (ls) \<and> (intlen ls) = n \<and> 
                          (nth (lsk) (intlen (lsk))) = ((intlen (suffix k \<sigma>))+k) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub ((nth lsk (i))) ((nth lsk ((i)+1))) (\<sigma>)) \<Turnstile> f)
                    ))
         )" 
 using interval_idx_link_shiftm by blast
 also have " ... =
            (\<exists> k.   0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls lsk. (intlen lsk) = n  \<and> index_sequence k (lsk) \<and>
                          lsk = map (shift k) ls \<and> 
                          index_sequence 0 (ls) \<and> (intlen ls) = n  \<and>
                          (nth (lsk) (intlen (lsk))) = ((intlen (suffix k \<sigma>))+k) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen lsk)) \<longrightarrow> 
                               ((sub ((nth lsk (i))) ((nth lsk ((i)+1))) (\<sigma>)) \<Turnstile> f)
                    ))
         )"
 using interval_lsk_ls by blast
 also have " ... =
            (\<exists> k ls lsk .   0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               ( (intlen lsk) = n  \<and> lsk = map (shift k) ls \<and> 
                           index_sequence 0 (ls) \<and>
                           index_sequence k (lsk) \<and>
                          (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub ((nth ls (i))+k) ((nth ls ((i)+1))+k) (\<sigma>)) \<Turnstile> f)
                    ))
         )"
 by (simp add: Interval.shift_def interval_nth_map, blast)
 also have " ... =
            (\<exists> k ls lsk.   0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               ( (intlen lsk) = n  \<and> lsk = map (shift k) ls \<and>
                          (intlen ls) = n  \<and> index_sequence 0 (ls) \<and> 
                          (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub ((nth ls (i))+k) ((nth ls ((i)+1))+k) (\<sigma>)) \<Turnstile> f)
                    ))
         )"
 using interval_idx_link by blast
 also have " ... =
            (\<exists> k. 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls. (intlen ls) = n  \<and> index_sequence 0 (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub ((nth ls (i))+k) ((nth ls ((i)+1))+k) (\<sigma>)) \<Turnstile> f)
                    ))
         ) "
 by (simp)
 also have " ... =
            (\<exists> k . 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls. (intlen ls) = n  \<and> index_sequence 0 (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and>
                    (\<forall>i\<le>intlen ls. Interval.nth ls i \<le> intlen (suffix k \<sigma>)) \<and>
                    (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub ((nth ls (i))+k) ((nth ls ((i)+1))+k) (\<sigma>)) \<Turnstile> f)
                    )
               )
         ) " 
 using interval_idx_bound_1 by blast
 also have " ... =
            (\<exists> k. 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls. (intlen ls) = n  \<and> index_sequence 0 (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and>
                   (\<forall>i\<le>intlen ls. Interval.nth ls i \<le> intlen (suffix k \<sigma>)) 
                 \<and> (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub (nth ls (i)) (nth ls ((i)+1)) (suffix k \<sigma>)) \<Turnstile> f)
                    )
               )
         ) " 
 by (smt add.commute index_sequence_def interval_idx_expand interval_sub_suffix 
         interval_suffix_length_good plus_1_eq_Suc)
 also have " ... =
            (\<exists> k. 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls. (intlen ls) = n  \<and> index_sequence 0 (ls) \<and> 
                      (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) 
                 \<and> (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                               ((sub (nth ls (i)) (nth ls ((i)+1)) (suffix k \<sigma>)) \<Turnstile> f)
                    ))
         ) " 
 using interval_idx_bound_1 by blast
 finally show "(\<exists> (l::index). (intlen l) = (Suc n) \<and> index_sequence 0 l \<and> 
                  (nth l (intlen l)) = (intlen \<sigma>) \<and>
                  (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                        ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
                  )
                ) =
                (\<exists> k. 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and> (sub 0 k \<sigma> \<Turnstile> f) \<and> 
                  (\<exists> ls. (intlen ls) = n \<and>index_sequence 0 (ls) \<and> 
                     (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and> 
                     (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                        ((sub (nth ls (i)) (nth ls ((i)+1)) (suffix k \<sigma>)) \<Turnstile> f)
                    )
                  )
                )
               " .
qed

lemma chop_power_eqv_sem:
  "( (\<sigma> \<Turnstile>(\<exists> n. (power (f \<and> more) n)))) = 
   ((\<sigma> \<Turnstile> empty) \<or>  ( (\<sigma> \<Turnstile> (f \<and> more);(\<exists> n. (power (f \<and> more) n)))))"
using ChopstarEqvSem powerstar_d_def chopstar_d_def
by (metis (mono_tags, lifting) unl_lift2)

lemma chopstar_eqv_power_chop_help:
  " ( \<sigma> \<Turnstile> power (f \<and> more) n) = 
    (\<exists> (l::index). intlen(l) = n \<and>  index_sequence 0 l  \<and> 
      (nth l (intlen l)) = (intlen ( \<sigma>)) \<and>
      (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
            ((sub (nth l i) (nth l (i+1)) (\<sigma>)) \<Turnstile> f)
      )
    ) "
proof 
 (induct n arbitrary: \<sigma>)
 case 0
 then show ?case using index_sequence_def chopstar_help_1 empty_defs 
 by (metis (mono_tags, lifting) intlen.simps(1) pow_0)
 next
 case (Suc n)
 then show ?case 
  proof -
   have 1: "(\<sigma> \<Turnstile> power (f \<and> more) (Suc n)) = (\<sigma> \<Turnstile> ((f \<and> more);(power (f \<and> more) n)))  " 
   by simp
   have 2: "(\<sigma> \<Turnstile> ((f \<and> more);(power (f \<and> more) n))) =
             (\<exists> k. 0 \<le>k \<and> k \<le> intlen (\<sigma>) \<and> k > 0 \<and>
               (prefix k (\<sigma>) \<Turnstile> f) \<and> 
               (suffix k (\<sigma>) \<Turnstile> power (f \<and> more) n)
             )
           " 
  by (simp add: more_defs chop_defs) auto
  have 3: "(\<exists> k. 0 \<le>k \<and> k \<le> intlen (\<sigma>) \<and> k > 0 \<and>
               (prefix k (\<sigma>) \<Turnstile> f) \<and> 
               (suffix k (\<sigma>) \<Turnstile> power (f \<and> more) n)
           ) =
           (\<exists> k. 0 \<le>k \<and> k \<le> intlen (\<sigma>) \<and> k > 0 \<and>
               (sub 0 k (\<sigma>) \<Turnstile> f) \<and> 
               (suffix k (\<sigma>) \<Turnstile> power (f \<and> more) n)
           )" 
  by (simp add: interval_sub_zero_prefix)
  have 31: "\<And> k. ((suffix k \<sigma>) \<Turnstile> power (f \<and> more) n) =
             (\<exists> (l::index). intlen(l) = n \<and>  index_sequence 0 l  \<and> 
                  (nth l (intlen l)) = (intlen (suffix k \<sigma>)) \<and>
                  (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                       ((sub (nth l i) (nth l (i+1)) (suffix k \<sigma>)) \<Turnstile> f)
                  )
                )"
    by (simp add: Suc.hyps)
  have 4: "(\<exists> k. 0 \<le>k \<and> k \<le> intlen (\<sigma>) \<and> k > 0 \<and>
               (sub 0 k (\<sigma>) \<Turnstile> f) \<and> 
               (suffix k (\<sigma>) \<Turnstile> power (f \<and> more) n)
           ) =
           (\<exists> k. 0 \<le>k \<and> k \<le> intlen (\<sigma>) \<and> k > 0 \<and>
               (sub 0 k (\<sigma>) \<Turnstile> f) \<and> 
               (\<exists> (l::index). intlen(l) = n \<and>  index_sequence 0 l  \<and> 
                  (nth l (intlen l)) = (intlen (suffix k \<sigma>)) \<and>
                  (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                       ((sub (nth l i) (nth l (i+1)) (suffix k \<sigma>)) \<Turnstile> f)
                  )
                )
             )" 
   using "31" by simp
   have 5: "
           (\<exists> (l::index). (intlen l) = (Suc n) \<and> index_sequence 0 l \<and> 
              (nth l (intlen l)) = (intlen \<sigma>) \<and>
              (\<forall>i. (0\<le>i \<and> i< (intlen l)) \<longrightarrow> 
                   ((sub (nth l i) (nth l (i+1)) \<sigma>) \<Turnstile> f)
              )
           ) =
           (\<exists> k. 0 \<le>k \<and> k \<le> intlen \<sigma> \<and> k > 0 \<and>
               (sub 0 k \<sigma> \<Turnstile> f) \<and> 
               (\<exists> ls. (intlen ls) = n \<and>index_sequence 0 (ls) \<and> 
                  (nth (ls) (intlen (ls))) = (intlen (suffix k \<sigma>)) \<and> 
                  (\<forall>i. (0\<le>i \<and> i< (intlen ls)) \<longrightarrow> 
                        ((sub (nth ls (i)) (nth ls ((i)+1)) (suffix k \<sigma>)) \<Turnstile> f)
                  )
               )
           )" 
    using chop_power_chain by simp
    from 1 2 3 4 5 show ?thesis by blast
  qed
qed

lemma chopstar_eqv_power_chop:
  "(\<sigma> \<Turnstile> chopstarold f) = (  (\<sigma> \<Turnstile> (\<exists> k. power (f \<and> more) k))) "
by (simp add: chopstar_d_old_def chopstar_eqv_power_chop_help)

lemma OldChopstarEqvSem:
  " (\<sigma> \<Turnstile> (chopstarold f =  (empty \<or>  (f \<and> more); (chopstarold f))) )"
using chopstar_eqv_power_chop chop_power_eqv_sem 
by (smt chop_defs unl_lift2)

lemma OldChopstarEqvChopstar:
 " \<turnstile> (chopstarold f) = f\<^sup>\<star>"
by (simp add: Valid_def chopstar_d_def chopstar_eqv_power_chop powerstar_d_def)


end
